import numpy as np

import EnemyCaseList
from EnemyCaseList import EnemyCaseList

import aiwolf, to_nlp
#import from_nlp
import from_nlp_batch as from_nlp
from template import talk_check, whisper_check

debug_mode = 'y'
def print_debug(x):
    if debug_mode == 'y':
        print(x)

def to_official_index(i):
    return i + 1

def argmax(lst):
    best_index = 0
    best_value = -float('inf')
    best_count = 1
    for i, v in enumerate(lst):
        if v > best_value:
            best_index = i
            best_value = v
            best_count = 1
        elif v >= best_value:
            if np.random.randint(0, best_count) == 0:
                best_index = i
            best_count += 1
    return best_index

def argmin(lst):
    best_index = 0
    best_value = float('inf')
    best_count = 1
    for i, v in enumerate(lst):
        if v < best_value:
            best_index = i
            best_value = v
            best_count = 1
        elif v <= best_value:
            if np.random.randint(0, best_count) == 0:
                best_index = i
            best_count += 1
    return best_index

class StrategyPlayer(object):
    
    def __init__(self, game_info, game_setting, nlp_mode):
        self.vilsize = game_setting['playerNum']
        self.nlp_mode = nlp_mode
        self.agentstringmap = {'-1':-1}
        self.agentstringlist = []
        self.agent_names = []
        for i, agent_str in enumerate(game_info['statusMap'].keys()):
            self.agentstringmap[agent_str] = i
            self.agentstringlist.append(agent_str)
            self.agent_names.append('Agent[' + ('%02d' % int(agent_str)) + ']')
        self.agentIdx = self.agentstringmap[str(game_info['agent'])]
        self.prob_ww = [0.0] * self.vilsize
        self.prob_ww_wn = [0.0] * self.vilsize
        self.role = game_info['roleMap'][str(game_info['agent'])]
        self.werewolves = []
        for agent_str in game_info['roleMap'].keys():
            if game_info['roleMap'][agent_str] == "WEREWOLF":
                self.werewolves.append(self.agentstringmap[agent_str])
        self.comingout = None
        self.not_reported = False
        self.divine_result = []
        self.identify_result = []
        self.cum_colist = []
        self.votelist = []
        self.colist = []
        self.divinelist = []
        self.identifylist = []
        self.status_list = ["ALIVE"] * self.vilsize
        self.day = 0
        self.vote_declare = False
        self.attack_declare = False
        self.eclist = EnemyCaseList(self.vilsize)
        self.eclist.initialize()
        self.divined_flag = [0] * self.vilsize
        self.opening = True
        print_debug('NEW GAME')
        print_debug(('My agent index = ', self.agentIdx))
        
    
    def update(self, game_info, talk_history):
        # print_debug(game_info)
        # find day
        if 'day' in game_info:
            self.day = game_info["day"]
        # find result
        # divine_result
        if self.role == 'SEER' and 'divineResult' in game_info:
            if game_info['divineResult'] is not None:
                if game_info['divineResult']["result"] == "HUMAN":
                    self.divine_result = [self.agentstringmap[str(game_info['divineResult']["target"])], 0]
                else:
                    self.divine_result = [self.agentstringmap[str(game_info['divineResult']["target"])], 1]
                self.divined_flag[self.agentstringmap[str(game_info['divineResult']["target"])]] = 1
                self.not_reported = True
        # identify_result
        elif self.role == 'MEDIUM' and 'mediumResult' in game_info:
            if game_info['mediumResult'] is not None:
                if game_info['mediumResult']["result"] == "HUMAN":
                    self.identify_result = [self.agentstringmap[str(game_info['mediumResult']["target"])], 0]
                else:
                    self.identify_result = [self.agentstringmap[str(game_info['mediumResult']["target"])], 1]
                self.not_reported = True
        # find attacked
        attacked_agent = -1 # -1 means none
        if 'attackedAgent' in game_info:
            attacked_agent = self.agentstringmap[str(game_info['attackedAgent'])]
            if attacked_agent != -1:
                if self.status_list[attacked_agent] == "DEAD":
                    attacked_agent = -1
        # update statusmap
        if 'statusMap' in game_info:
            for istr, status in game_info['statusMap'].items():
                self.status_list[self.agentstringmap[istr]] = status
        # find votelist
        if 'voteList' in game_info:
            if len(game_info['voteList']) > 0:
                self.votelist = [[self.agentstringmap[str(vote_json['agent'])], self.agentstringmap[str(vote_json['target'])]]
                                 for vote_json in game_info['voteList'] if self.agentstringmap[str(vote_json['target'])] >= 0]
        # find comingout, divined, identifyed, guarded from talk
        if len(talk_history) > 0:
            texts = []
            subjects = []
            for talk_json in talk_history:
                texts.append(talk_json['text'])
                subjects.append(self.agentstringmap[str(talk_json['agent'])])
            if self.nlp_mode:
                print(texts,subjects)
                texts = from_nlp.from_japanese(texts, subjects, self.agent_names)
            for i, text in enumerate(texts):
                content = text.split()
                subject = subjects[i]
                # COMINGOUT
                if content[0] == 'COMINGOUT':
                    if content[2] == 'SEER':
                        co_role_no = 0
                    elif content[2] == 'MEDIUM':
                        co_role_no = 1
                    elif content[2] == 'BODYGUARD':
                        co_role_no = 2
                    else:
                        # ignore werewolf_co etc
                        co_role_no = 9
                    newco = [subject, co_role_no]
                    # check if it is new
                    if newco not in self.cum_colist and newco[1] != 9:
                        self.cum_colist.append(newco)
                        self.colist.append(newco)
                # DIVINED
                elif content[0] == 'DIVINED':
                    # REGARD CO SEER
                    newco = [subject, 0]
                    if newco not in self.cum_colist and newco[1] != 9:
                        self.cum_colist.append(newco)
                        self.colist.append(newco)
                    # new divine
                    if content[2] == 'WEREWOLF' and content[1] in self.agent_names:
                        self.divinelist.append([subject, self.agent_names.index(content[1]), 1])
                    elif content[1] in self.agent_names:
                        self.divinelist.append([subject, self.agent_names.index(content[1]), 0])
                # IDENTIFIED
                elif content[0] == 'IDENTIFIED':
                    # REGARD CO MEDIUM
                    newco = [subject, 1]
                    if newco not in self.cum_colist and newco[1] != 9:
                        self.cum_colist.append(newco)
                        self.colist.append(newco)
                    # new identify
                    if content[2] == 'WEREWOLF' and content[1] in self.agent_names:
                        self.identifylist.append([subject, self.agent_names.index(content[1]), 1])
                    elif content[1] in self.agent_names:
                        self.identifylist.append([subject, self.agent_names.index(content[1]), 0])
                # GUARDED
                elif content[0] == 'GUARDED':
                    # REGARD CO BODYGUARD
                    newco = [subject, 2]
                    if newco not in self.cum_colist and newco[1] != 9:
                        self.cum_colist.append(newco)
                        self.colist.append(newco)
        #  what to do
        if attacked_agent != -1:
            print_debug('agent ' + str(attacked_agent) + 'is attacked')
            self.eclist.get_attacked(attacked_agent)
            print_debug('get votelist')
            print_debug(self.votelist)
            self.eclist.get_votelist(self.votelist)
            self.vote_declare = False
            self.votelist = []
        elif len(self.colist) > 0:
            print_debug('get colist')
            print_debug(self.colist)
            self.eclist.get_colist(self.colist)
            self.vote_declare = False
            self.colist = []
        elif len(self.divinelist) > 0:
            print_debug('get divinelist')
            print_debug(self.divinelist)
            self.eclist.get_divine_list(self.divinelist)
            self.vote_declare = False
            self.divinelist = []
        elif len(self.identifylist) > 0:
            print_debug('get identifylist')
            print_debug(self.identifylist)
            self.eclist.get_identify_list(self.identifylist)
            self.vote_declare = False
            self.identifylist = []
        else:
            # print_debug('Over')
            pass
        # due to 0.1s limit, only 1 action
        return None
    
    def dayStart(self):
        # possesed lie
        if self.role == 'POSSESSED' and self.day > 0:
            # who is most likely werewolf
            self.eclist.update_rel_prob()
            for i in range(0, self.vilsize):
                self.prob_ww_wn[i] = self.eclist.prob_now_wn(i, self.agentIdx)[2]
            print_debug(self.prob_ww_wn)
            # do not divine dead agent
            for i in range(0, self.vilsize):
                if self.status_list[i] == "DEAD":
                    self.prob_ww_wn[i] = 0.0
            # do not divine self(not nesessary)
            self.prob_ww_wn[self.agentIdx] = 0.0
            # do not divine again
            for i in range(0, self.vilsize):
                if self.divined_flag[i] == 1:
                    self.prob_ww_wn[i] = 0.0
            divine_target = argmax(self.prob_ww_wn)
            self.divine_result = [divine_target, 0]
            self.not_reported = True
        # all roles
        # vote_declare flag
        self.vote_declare = False
        # attack_declare flag
        self.attack_declare = False
        return None
    
    def finish(self):
        return None
    
    def vote(self):
        # who is most likely werewolf
        self.eclist.update_rel_prob()
        for i in range(0, self.vilsize):
            self.prob_ww_wn[i] = self.eclist.prob_now_wn(i, self.agentIdx)[2]
        print_debug(self.prob_ww_wn)
        # do not vote dead agent
        for i in range(0, self.vilsize):
            if self.status_list[i] == "DEAD":
                self.prob_ww_wn[i] = 0.0
        # do not vote self(not nesessary)
        self.prob_ww_wn[self.agentIdx] = 0.0
        # werewolves do not vote werewolves
        for i in range(0, self.vilsize):
            if i in self.werewolves:
                self.prob_ww_wn[i] /= 2.0
        return int(self.agentstringlist[argmax(self.prob_ww_wn)])
    
    def attack(self):
        # who is least likely to be werewolf
        self.eclist.update_rel_prob()
        for i in range(0, self.vilsize):
            self.prob_ww[i] = self.eclist.prob_now(i)[2]
        print_debug(self.prob_ww)
        # do not attack dead agent
        for i in range(0, self.vilsize):
            if self.status_list[i] == "DEAD":
                self.prob_ww[i] = 1.0
        # do not attack werewolves
        for i in range(0, self.vilsize):
            if i in self.werewolves:
                self.prob_ww[i] = 1.0
        return int(self.agentstringlist[argmin(self.prob_ww)])
    
    def guard(self):
        # who is least likely to be werewolf
        self.eclist.update_rel_prob()
        for i in range(0, self.vilsize):
            self.prob_ww_wn[i] = self.eclist.prob_now_wn(i, self.agentIdx)[2]
        print_debug(self.prob_ww_wn)
        # do not guard dead agent
        for i in range(0, self.vilsize):
            if self.status_list[i] == "DEAD":
                self.prob_ww_wn[i] = 1.0
        # do not guard self
        self.prob_ww_wn[self.agentIdx] = 1.0
        return int(self.agentstringlist[argmin(self.prob_ww_wn)])
    
    def divine(self):
        # who is most likely werewolf
        self.eclist.update_rel_prob()
        for i in range(0, self.vilsize):
            self.prob_ww_wn[i] = self.eclist.prob_now_wn(i, self.agentIdx)[2]
        print_debug(self.prob_ww_wn)
        # do not divine dead agent
        for i in range(0, self.vilsize):
            if self.status_list[i] == "DEAD":
                self.prob_ww_wn[i] = 0.0
        # do not divine self(not nesessary)
        self.prob_ww_wn[self.agentIdx] = 0.0
        # do not divine again
        for i in range(0, self.vilsize):
            if self.divined_flag[i] == 1:
                self.prob_ww_wn[i] = 0.0
        return int(self.agentstringlist[argmax(self.prob_ww_wn)])

    def talk_protocol(self):
        # comingout anyway
        if self.role == 'SEER' and self.comingout is None:
            self.comingout = 'SEER'
            return 'COMINGOUT ' + self.agent_names[self.agentIdx] + ' SEER'
        elif self.role == 'MEDIUM' and self.comingout is None:
            self.comingout = 'MEDIUM'
            return 'COMINGOUT ' + self.agent_names[self.agentIdx] + ' MEDIUM'
        elif self.role == 'POSSESSED' and self.comingout is None:
            self.comingout = 'SEER'
            return 'COMINGOUT ' + self.agent_names[self.agentIdx] + ' SEER'
        # report
        if self.role == 'SEER' and self.not_reported:
            self.not_reported = False
            if self.divine_result[1] == 0:
                return 'DIVINED ' + self.agent_names[self.divine_result[0]] + ' HUMAN'
            else:
                return 'DIVINED ' + self.agent_names[self.divine_result[0]] + ' WEREWOLF'
        elif self.role == 'MEDIUM' and self.not_reported:
            self.not_reported = False
            if self.identify_result[1] == 0:
                return 'IDENTIFIED ' + self.agent_names[self.identify_result[0]] + ' HUMAN'
            else:
                return 'IDENTIFIED ' + self.agent_names[self.identify_result[0]] + ' WEREWOLF'
        elif self.role == 'POSSESSED' and self.not_reported:
            self.not_reported = False
            if self.divine_result[1] == 0:
                return 'DIVINED ' + self.agent_names[self.divine_result[0]] + ' HUMAN'
            else:
                return 'DIVINED ' + self.agent_names[self.divine_result[0]] + ' WEREWOLF'
        # not thought enough
        elif len(self.colist) > 0 or len(self.divinelist) > 0 or len(self.identifylist) > 0:
            return 'Skip'
        elif self.vote_declare == False:
            # vote_declare
            self.vote_declare = True
            self.eclist.update_rel_prob()
            for i in range(0, self.vilsize):
                self.prob_ww_wn[i] = self.eclist.prob_now_wn(i, self.agentIdx)[2]
            print_debug(self.prob_ww_wn)
            # do not vote dead agent
            for i in range(0, self.vilsize):
                if self.status_list[i] == "DEAD":
                    self.prob_ww_wn[i] = 0.0
            # do not vote self(not nesessary)
            self.prob_ww_wn[self.agentIdx] = 0.0
            # werewolves do not vote werewolves
            for i in range(0, self.vilsize):
                if i in self.werewolves:
                    self.prob_ww_wn[i] /= 2.0
            return 'VOTE ' + self.agent_names[argmax(self.prob_ww_wn)]
        else:
            return 'Over'
    
    def whisper_protocol(self):
        # not thought enough
        if len(self.colist) > 0 or len(self.divinelist) > 0 or len(self.identifylist) > 0:
            return 'Skip'
        elif self.attack_declare == False:
            # attack_declare
            self.attack_declare = True
            # who is least likely to be werewolf
            self.eclist.update_rel_prob()
            for i in range(0, self.vilsize):
                self.prob_ww[i] = self.eclist.prob_now(i)[2]
            print_debug(self.prob_ww)
            # do not attack dead agent
            for i in range(0, self.vilsize):
                if self.status_list[i] == "DEAD":
                    self.prob_ww[i] = 1.0
            # do not attack werewolves
            for i in range(0, self.vilsize):
                if i in self.werewolves:
                    self.prob_ww[i] = 1.0
            return 'ATTACK ' + self.agent_names[argmin(self.prob_ww)]
        else:
            return 'Over'

    def talk(self):
        #return 'Skip'
        if self.nlp_mode and self.day == 0:
            if self.opening:
                self.opening = False
                return to_nlp.to_opening_japanese(self.agentIdx)
            else:
                return 'Over'
        str = talk_check(self.talk_protocol(), self.agent_names)
        if self.nlp_mode:
            text = to_nlp.to_japanese(str, self.agentIdx)
            print_debug(text)
            return text
        else:
            return str

    def whisper(self):
        str = whisper_check(self.whisper_protocol(), self.agent_names)
        if self.nlp_mode:
            return to_nlp.to_japanese(str)
        else:
            return str

    def getName(self):
        return self.name

