# -*- coding: utf-8 -*-
# from_nlp_batch.py
# Katsuki Ohto

import sys, time

from yandex_translate import YandexTranslate

import aiwolf
from template import talk_check, whisper_check
import salmon#, d2v

sentences_en = []
sentences_jp = []
api = None

name_string_en = ['Alice', 'Bob', 'Charlie', 'Dave', 'Ellen']
                  #'Frank', 'Eve', 'Issac', 'Ivan', 'Justin',
                  #'Justin', 'Mallory', 'Matilda', 'Oscar', 'Pat']

name_string_jp = ['不二夫', '芳伸', '孝祐', '通将', '恭士']

pseudo_role_en = {'villagers':aiwolf.VILLAGER, 'psychics':aiwolf.SEER,
                  'madman':aiwolf.POSSESSED, 'wolf':aiwolf.WEREWOLF}
pseudo_species_en = {'human':aiwolf.HUMAN, 'wolf':aiwolf.T_WEREWOLF}

pseudo_role_jp = {'村人':aiwolf.VILLAGER, '占い師':aiwolf.SEER,
                  '狂人':aiwolf.POSSESSED, '狼':aiwolf.WEREWOLF}
pseudo_species_jp = {'人間':aiwolf.HUMAN, '人狼':aiwolf.T_WEREWOLF}

# aiwolf_stringへの登録
for i, name in enumerate(name_string_en):
    for sp, spi in pseudo_species_en.items():
        sentences_en.append((name + ' was ' + sp + '.', [-1, i], 'DIVINED ' + '*1 ' + aiwolf.SPECIES_NAMES[spi]))
    for ro, roi in pseudo_role_en.items():
        sentences_en.append((name + ' seems to be ' + ro + '.', [-1, i], 'ESTIMATE ' + '*1 ' + aiwolf.ROLE_NAMES[roi]))
    sentences_en.append((name + ' to vote.', [-1, i], 'VOTE ' + '*1'))

for sp, spi in pseudo_species_en.items():
    sentences_en.append(('You are ' + sp + '.', [-2], 'DIVINED ' + '*0 ' + aiwolf.SPECIES_NAMES[spi]))
for ro, roi in pseudo_role_en.items():
    sentences_en.append(('I am ' + ro + '.', [-1], 'COMINGOUT ' + '*0 ' + aiwolf.ROLE_NAMES[roi]))

for i, name in enumerate(name_string_jp):
    for sp, spi in pseudo_species_jp.items():
        sentences_jp.append((name + 'は' + sp + 'だった', [-1, i], 'DIVINED ' + '*1 ' + aiwolf.SPECIES_NAMES[spi]))
    for ro, roi in pseudo_role_jp.items():
        sentences_jp.append((name + 'は' + ro + 'だと思う', [-1, i], 'ESTIMATE ' + '*1 ' + aiwolf.ROLE_NAMES[roi]))
    sentences_jp.append((name + 'に投票する', [-1, i], 'VOTE ' + '*1'))

for sp, spi in pseudo_species_jp.items():
    sentences_jp.append(('あなたは' + sp + 'ですね', [-2], 'DIVINED ' + '*0 ' + aiwolf.SPECIES_NAMES[spi]))
for ro, roi in pseudo_role_jp.items():
    sentences_jp.append(('私は' + ro + 'です', [-1], 'COMINGOUT ' + '*0 ' + aiwolf.ROLE_NAMES[roi]))

print(sentences_en)
print(sentences_jp)

def initialize(api_key):
    global api
    api = YandexTranslate(api_key)
    print('initialized api connection')

def preprocess(texts, subjects, agent_string_list, name_string):
    # 前処理
    atexts = []
    objects = []
    agent_lists = []
    
    for i, text in enumerate(texts):
        
        if text == 'Skip' or text == 'Over':
            atexts.append(text)
            objects.append(None)
            agent_lists.append([])
            continue
        
        subject = subjects[i]
    
        object = None
        agent_list = []
    
        # アンカー処理
        if text[:2] == '>>':
            slist = text.split(' ')
            text = ''.join(slist[1:])
            anchor = slist[2:]
            for i, a in enumerate(agent_string_list):
                if a == anchor:
                    object = i
                    agent_list.append(-2)
                    break
        else:
            agent_list.append(-1)

        # 自分自身のエージェント名を一人称に変換
        subject_agent = agent_string_list[subject]
        if subject_agent in text:
            text = text.replace(subject_agent, '私')

        # エージェント名を人名に変換
        for i, agent in enumerate(agent_string_list):
            if agent in text:
                agent_list.append(i)
                text = text.replace(agent, name_string[i])

        # 役職名を変換
        text = text.replace('人狼', '狼')
        # 種族名を変換
    
        atexts.append(text)
        objects.append(object)
        agent_lists.append(agent_list)
    
    return atexts, agent_lists, objects

def translate(texts, api):
    return api.translate(texts, 'ja-en')['text']

def postprocess(text, marks, subject, object, agent_string_list):
    # 後処理
    for j, a in enumerate(marks):
        mark = '*' + str(j)
        if mark in text:
            if a == -1:
                if subject is not None:
                    text = text.replace(mark, agent_string_list[subject])
            elif a == -2:
                if object is not None:
                    text = text.replace(mark, agent_string_list[object])
            else:
                text = text.replace(mark, agent_string_list[a])
    text = talk_check(text, agent_string_list)
    return text

def from_japanese_en(otexts, subjects, agent_string_list):
    
    texts, alists, objects = preprocess(otexts, subjects, agent_string_list, name_string_en)
    print('after preprocess : ', texts, alists, objects)
    texts = translate(texts, api)
    print('after translation : ', texts)
    
    answers = []
    # 文章間距離を計算
    for i, text in enumerate(texts):
        if otexts[i] == 'Skip' or otexts[i] == 'Over':
            answers.append(otexts[i])
            continue
        
        subject = subjects[i]
        object = objects[i]
        alist = alists[i]
        lst = []
        for i, sen in enumerate(sentences_en):
            stext, alst, ans = sen
            if alist == alst:
                dist = salmon.similarity(stext, text, True)
                lst.append((i, dist))
        if len(lst) == 0 or lst[0][1] < 0.05:
            answers.append('Skip')
        else:
            if len(lst) > 1:
                lst = sorted(lst, key = lambda x:-x[1])

            text = sentences_en[lst[0][0]][2]
            marks = sentences_en[lst[0][0]][1]
            text = postprocess(text, marks, subject, object, agent_string_list)

            answers.append(text)
    print('after postprocess : ', answers)
    
    return answers

def from_japanese_jp(text, subject, agent_string_list):

    if text == 'Skip' or text == 'Over':
        return text
    
    text, alist, object = preprocess(text, subject, agent_string_list, name_string_jp)
    #print('after preprocess : ', text, alist, object)
    
    # 文章間距離を計算
    lst = []
    for i, sen in enumerate(sentences_jp):
        stext, alst, ans = sen
        if alist == alst:
            dist = d2v.similarity(stext, text)
            lst.append((i, dist))
    if len(lst) == 0 or lst[0][1] < 0.1:
        return 'Skip'
    elif len(lst) > 1:
        lst = sorted(lst, key = lambda x:-x[1])
    
    text = sentences_jp[lst[0][0]][2]
    marks = sentences_jp[lst[0][0]][1]
    text = postprocess(text, marks, subject, object, agent_string_list)
    #print('after postprocess : ', text)
    return text

def from_japanese(texts, subjects, agent_string_list):
    return from_japanese_en(texts, subjects, agent_string_list)

if __name__ == '__main__':
    
    initialize(sys.argv[1])
    asl = ['Agent[' + ('%02d' % i) + ']' for i in range(1, 5+1)]
    print(asl)
    print(from_japanese(['私は人狼です'], [0], asl))
    
    # 翻訳計算
    sample_texts = ['Agent[01]は人狼でした',
                    'Agent[01]は人間でした',
                    'Agent[02]は人狼でした',
                    'Agent[02]は人間でした',
                    '私は村人です',
                    '私は狂人です',
                    '私は占い師です',
                    '私は人狼です',
                    '>>Agent[01] あなたは人狼ですね']
        
    t = time.time()
    atexts = from_japanese(sample_texts, [0] * len(sample_texts), asl)
    print(atexts)
    print(time.time() - t)


